var fs = require('fs');
var path = require('path');
var singularize = require("../vendor/inflection").singularize;
var express = require('express');
var utils = require('./sense_utils');

var safe_merge = utils.safe_merge;
var Mapper = require('./route_mapper').Mapper;
require('coffee-script');

function Sense() {
  global.sense = this;

  this.locales     = require('./locales');
  this.utils       = require('./sense_utils');

  this.ControllerBridge = require('./controller_bridge');
  this.ControllerBridge.init();

  this.controller  = require('./controller');
  this.extensions  = require('./extensions');
  this.generators  = require('./generators');
  this.tools       = require('./tools');
  this.logger      = require('./logger');
  this.routeMapper = new Mapper(app);
  this.helpers     = require('./helpers');

  this.models      = require('./models');
}

exports.init = function (app) {
  if (arguments.length == 2) {
    app = arguments[1];
  }

  // globalize app object
  global.app = app;
  app.root = process.cwd();
  // create API publishing object
  new Sense();

  // run environment.{js|coffee} and environments/{test|development|production}.{js|coffee}
  configureApp();

  hookSockets();

  // controllers should be loaded before extensions
  sense.controller.init();
  // extensions should be loaded before server startup
  sense.extensions.init();

  sense.models.init();

  // run config/initializers/*
  runInitializers();

  if (path.existsSync(app.root + '/config') && (path.existsSync(app.root + '/config/routes.js') || path.existsSync(app.root + '/config/routes.coffee'))) {
    sense.routeMapper.addRoutes(app.root + '/config/routes');
  }

  //if (process.env.JS_SOCKETS === 'on') {
    //app.on('listening', hookSockets);
  //}

  process.nextTick(function () {

    sense.locales.init();
    sense.logger.init();
    app.reloadModels = sense.models.loadModels;

    loadObservers();

    if (app.enabled('merge javascripts')) {
      ensureDirClean(app.root + '/public' + app.set('jsDirectory') + 'cache');
    }

    if (app.enabled('merge stylesheets')) {
      ensureDirClean(app.root + '/public' + app.set('cssDirectory') + 'cache');
    }

  });
};

exports.createServer = function (options) {
  options = options || {};

  var keys, app,
      key = options.key || process.cwd() + '/config/tsl.key',
      cert = options.cert || process.cwd() + '/config/tsl.cert';

  if (path.existsSync(key) && path.existsSync(cert)) {
    keys = {
     key: fs.readFileSync(key).toString('utf8'),
     cert: fs.readFileSync(cert).toString('utf8')
    };
  }

  if (keys) {
    app = express.createServer(keys);
  } else {
    app = express.createServer();
  }

  exports.init(app);
  return app;
};

function configureApp() {
  var mainEnv = app.root + '/config/environment';
  requireIfExists(mainEnv + '.js') || requireIfExists(mainEnv + '.coffee');
  var supportEnv = app.root + '/config/environments/' + app.settings.env;
  requireIfExists(supportEnv + '.js') || requireIfExists(supportEnv + '.coffee');

  if (app.settings['view engine'] == 'ejs' && (!app.extensions || !app.extensions['ejs-extension'])) {
    // monkey patch ejs
    var ejs = require('ejs'), old_parse = ejs.parse;
    ejs.parse = function () {
      var str = old_parse.apply(this, Array.prototype.slice.call(arguments));
      return str.replace('var buf = [];', 'var buf = []; arguments.callee.buf = buf;');
    };
  }

  if (app.settings['view engine'] == 'jade' && (!app.extensions || !app.extensions['jade-extension'])) {
    // monkey patch jade
    var jade = require('jade'), old_parse = jade.Compiler.prototype.compile;
    jade.Compiler.prototype.compile = function () {
      var str = old_parse.apply(this, Array.prototype.slice.call(arguments));
      return 'arguments.callee.buf = buf;' + str;
    };
  }
}

function requireIfExists(module) {
  if (path.existsSync(module)) {
    require(module);
    return true;
  } else {
    return false;
  }
}

function runInitializers() {
  var context = {global: {}};

  for (var i in app.models) {
    context[i] = app.models[i];
  }

  var initializersPath = app.root + '/config/initializers/';
  if (require('path').existsSync(initializersPath)) {
    fs.readdirSync(initializersPath).forEach(function (file) {
      if (file.match(/^\./)) return;
      var script_name = initializersPath + file;
      utils.runCode(script_name, context);
    });
    for (var i in context.global) {
      global[i] = context.global[i];
    }
  }
  app.global = context.global;
}

/**
 * Observer is a kind of controller, that listen for some event in
 * the system, for example: paypal, twitter or facebook observers
 * listens for callback from foreign service. Email observer may
 * listen some events related to emails.
 *
 * If you need app.on('someEvent') you should place this code in
 * APPROOT/app/observers/NAME_observer.js
 */
function loadObservers() {
  var dir = app.root + '/app/observers';
  path.exists(dir, function (exists) {
    if (exists) {
      fs.readdir(dir, function (err, files) {
        if (!err && files) {
          files.forEach(function (file) {
            require(path.join(dir, file));
          });
        }
      });
    }
  });
}

function ensureDirClean(dir) {
  path.exists(dir, function (exists) {
    if (exists) {
      fs.readdir(dir, function (err, files) {
        files.map(function (file) {
          return path.join(dir, file);
        }).forEach(fs.unlink);
      });
    } else {
      fs.mkdir(dir, 0755);
    }
  });
}

function hookSockets() {
  var io = app.Socket = require('socket.io').listen(app);
  io.configure('development', function(){
    io.set('transports', ['websocket', 'flashsocket', 'htmlfile', 'xhr-polling', 'jsonp-polling']);
    io.enable('browser client etag');
    io.set('log level', 2);
  });
  /*
  message = {
    c: 'foursquare',
    a: 'teste',
    m: {teste: 'ssss'}
  }
  */
  io.sockets.on('connection', function (client) {
    app.SC = client;
    client.on('request', function (m) {
      var ctl = sense.controller.load(m.c, m.c, '');
      ctl.performSocket(client, m);
    });

    client.on('disconnect', function () {
        console.warn('disconnecting');
    });

  });
}